/*
 * (c) 2023-2024 Jens Mueller
 *
 * JKCLOAD
 *
 * Hex-ASCII-Anzeige des Dateiinhalts
 */

package jkcload.ui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.util.Properties;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import jkcload.Main;


public class HexAsciiFrm extends JFrame implements ActionListener
{
  private static final String PROP_PREFIX = Main.PROP_PREFIX + "hexascii.";

  private boolean   notified;
  private JMenuItem mnuClose;
  private JTextArea textArea;


  public HexAsciiFrm()
  {
    this.notified = false;
    setDefaultCloseOperation( JFrame.HIDE_ON_CLOSE );
    setTitle( "HEX-ASCII-Anzeige" );


    // Fenster-Icons
    UIUtil.setIconImagesAt( this );


    // Menu
    JMenu mnuFile = new JMenu( "Datei" );
    mnuFile.setMnemonic( KeyEvent.VK_D );

    this.mnuClose = new JMenuItem( "Schlie\u00DFen" );
    mnuFile.add( this.mnuClose );

    JMenuBar mnuBar = new JMenuBar();
    mnuBar.add( mnuFile );
    setJMenuBar( mnuBar );


    // Fensterinhalt
    setLayout( new BorderLayout() );
    this.textArea = new JTextArea( 20, 75 );
    this.textArea.setEditable( false );
    add( new JScrollPane( this.textArea ), BorderLayout.CENTER );


    // Sonstiges
    if( !UIUtil.restoreWindowBounds(
			this,
			Main.getProperties(),
			PROP_PREFIX ) )
    {
      pack();
      setLocationByPlatform( true );
    }
  }


  public void getSettings( Properties props )
  {
    UIUtil.getWindowBounds( this, props, PROP_PREFIX );
  }


  public void setContentBytes( byte[] contentBytes )
  {
    String text = "";
    if( contentBytes != null ) {
      StringBuilder buf = new StringBuilder(
				((contentBytes.length + 15) / 16) * 72 );
      int offs = 0;
      while( offs < contentBytes.length ) {
	buf.append( String.format( "%04X ", offs ) );
	for( int i = 0; i < 16; i++ ) {
	  int idx = offs + i;
	  if( idx < contentBytes.length ) {
	    buf.append( String.format(
				" %02X",
				contentBytes[ idx ] & 0xFF ) );
	  } else {
	    buf.append( "   " );
	  }
	}
	buf.append( "  " );
	for( int i = 0; i < 16; i++ ) {
	  int idx = offs + i;
	  if( idx < contentBytes.length ) {
	    int b = contentBytes[ idx ] & 0xFF;
	    buf.append( (char) ((b >= 0x20) && (b < 0x7F) ? b : '.') );
	  } else {
	    buf.append( ' ' );
	  }
	}
	buf.append( '\n' );
	offs += 16;
      }
      text = buf.toString();
    }
    this.textArea.setText( text );
    try {
      this.textArea.setCaretPosition( 0 );
    }
    catch( IllegalArgumentException ex ) {}
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.mnuClose ) {
      setVisible( false );
    }
  }


	/* --- ueberschriebene Methoden --- */

  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      this.mnuClose.addActionListener( this );
    }
  }


  @Override
  public void removeNotify()
  {
    super.removeNotify();
    if( this.notified ) {
      this.notified = false;
      this.mnuClose.removeActionListener( this );
    }
  }
}
